declare name "piano";
declare description "WaveGuide Commuted Piano";
declare author "Romain Michon (rmichon@ccrma.stanford.edu)";
declare author "Punk Labs LLC";
declare copyright "Romain Michon";
declare copyright "2024 Punk Labs LLC";
declare version "1.0";
declare licence "STK-4.3"; // Synthesis Tool Kit 4.3 (MIT style license);
declare description "A commuted WaveGuide piano. Modified for use in OneTrick Keys"; 

import("instruments.lib");
import("piano.h.lib");
import("onetrick.lib");
import("params.lib");

//==================== ADJUSTED FOR SAMPLERATE ================
// Delay-based fixes to make these filters samplerate independent...

TO_SR = ot.CD_TO_SR; // Values were originally designed for 44.1khz

//onePoleSR = onePole;
onePoleSR(b0,a1,x) = (b0*x - a1*_last)~_ with {
	_last = (_:de.fdelay(16, max(1, TO_SR)-1));
};

// poleZeroSR = poleZero;
poleZeroSR(b0,b1,a1,x) = (b0*x + b1*x_last - a1*_last)~_ with {
	x_last = x:de.fdelay(16, max(1, TO_SR));
	_last = (_:de.fdelay(16, max(1, TO_SR)-1));
};

// Not actually used?
// oneZero1SR(b0,b1,x) = (x_last*b1 + x*b0) with {
// 	x_last = x:de.fdelay(16, max(1, TO_SR));
// };

//==================== GUI SPECIFICATION ================


// freq: Use for simpler Control Rate calculations
freq = ot.input.freq*ot.input.pitchShift(pitchWheelRange);
// freqSmooth: Use for more costly Sample Rate calculations
freqSmooth = ot.input.freq*ot.input.pitchShiftSmooth(pitchWheelRange);

gain = ot.input.gain;
gate = ot.input.gate;

brightnessFactor = brightness * 0.25 - 0.05;
tuningRange = tuningFactor * 0.25 : ba.semi2ratio;
stiffnessFactor = 0.5; // String Stiffness: Doesn't contribute enough, can become unstable... Max Safe: 0.8
hammerHardness = 0.2; // Really a kind of sustain, we'll take max (0-0.1-1.0)

//==================== COMMUTED PIANO PARAMETERS ================

//convert a frequency in a midi note number
freqToNoteNumber = ba.hz2midikey;//(log-log(440))/log(2)*12+69+0.5 : int;
freqToPianoNoteNumber = ba.hz2pianokey;//(log-log(440))/log(2)*12+69+0.5 : int;
note = freq : freqToNoteNumber;
pianoNote = freq : freqToPianoNoteNumber;
note_01 = note:it.remap(21, 109, 0, 1);


//a counter that restart a every note-on
cntSample = *(gate)+1~_ : -(1);

//==================== PIANO SOUND BOARD ================

//exponential envelope with 3 phases for the pedal excitation	
asympT60pedal(value,T60) = (*(factor) + constant)~_
	with{
		attDur = hammerHardness*float(ma.SR);
		target = value*((cntSample < attDur) & (gate > 0));
		factorAtt = exp (-1/(attDur)); 
		factorG = exp(-1/(2*float(ma.SR)));
		factorT60 = exp(-7/(T60*float(ma.SR)));
		factor = factorAtt*gate*(cntSample < attDur) + (cntSample >= attDur)*gate*factorG + ((gate-1)*-1)*factorT60;
		constant = (1 - factor)*target;			
	};

//the sound of the piano sound board is generated by noise generator whose output gain is shaped by
//an exponential envelope
soundBoard = dryTapAmp*ot.noise + pedalEnv*ot.noise : *(0.5)
	with{
		dryTapAmpT60 = getValueDryTapAmpT60; // <piano.h>
		sustainPedalLevel = getValueSustainPedalLevel; // <piano.h>

		pedalEnvCutOffTime = 1.4;
		noteCutOffTime = note : dryTapAmpT60*gain;
		pedalEnvValue = note : sustainPedalLevel*0.2;
		noteEnvValue = 0.15;
		dryTapAmp = asympT60(noteEnvValue,0,noteCutOffTime,gate);
		pedalEnv = asympT60pedal(pedalEnvValue,pedalEnvCutOffTime);
	};	

//==================== HAMMER MODELING ================

//To model the exitation hammer, we filter the sound from the soundboard with a serie of 4 one pole filters
//connected in serie  

//onePole is declared in instruments.lib
calcHammer = onePoleSR((1-hammerPole)*hammerGain,-hammerPole)
	with{
		loudPole = getValueLoudPole; // <piano.h>
		softPole = getValuePoleValue; // <piano.h>
		loudGain = getValueLoudGain; // <piano.h>
		softGain = getValueSoftGain; // <piano.h>

		loudPoleValue = loudPole(note) + (brightnessFactor*-0.25) + 0.02;
		softPoleValue = softPole(note);
		normalizedVelocityValue = gain; //pow(gain, 1.6);
		loudGainValue = loudGain(note);
		softGainValue = loudGainValue;//softGain(note);
		overallGain = 1; //2-gain;
		hammerPole = softPoleValue + (loudPoleValue - softPoleValue)*normalizedVelocityValue;
		hammerGain = overallGain*(softGainValue + (loudGainValue - softGainValue)*normalizedVelocityValue);
	};

hammer = seq(i,4,calcHammer); //5-pole is softer than original 4-pole...

//==================== DC BLOCKERS ================

// DCBa1 = getValueDCBa1; // <piano.h>
// dCBa1Value = note : DCBa1;
// dCBb0Value = 1 - dCBa1Value;
// dcBlock1 = poleZeroSR((dCBb0Value*0.5),(dCBb0Value*-0.5),dCBa1Value);
dcBlock1 = fi.highpass(1, freq);

// This doesn't work because dcBlock1 is used more like a highpass filter...
// dcBlock1 = ot.fi.dcblocker;
	
//==================== STRIKE POSITION COMB FILTER EQ ================

eq = _*filterGain : fi.tf2(b0,b1,b2,a1,a2)
	with{
		strikePosition = getValueStrikePosition; // <piano.h>
		bandwidthFactors = getValueEQBandWidthFactor; // <piano.h>
		eq_gain = getValueEQGain; // <piano.h>
		eq_tuning = freq/strikePosition(note);
		eq_bandwidth = bandwidthFactors(note)*freq;
		filterGain = eq_gain(note);
		a2 = (eq_bandwidth / ma.SR) * (eq_bandwidth / ma.SR);
		a1 = -2*(eq_bandwidth / ma.SR)*cos(2*ma.PI*eq_tuning/ma.SR);
		b0 = 0.5 - 0.5 * a2;
		b1 = 0;
		b2 = -b0;	
	};
	
//==================== PIANO COUPLED STRINGS ================

singleStringDecRate = getValueSingleStringDecayRate; // <piano.h>
singleStringZero = getValueSingleStringZero; // <piano.h>
singleStringPole = getValueSingleStringPole; // <piano.h>
stiffnessCoefficient = getValueStiffnessCoefficient; // <piano.h>

//coupling filter parameters
//enhanceSustain = 50/100;
enhanceSustain = it.interpolate_linear(bodyDamp, 40/100, 0);
g = it.interpolate_linear(enhanceSustain, pow(10,((singleStringDecRate(note)/freq)/20)), 1); //attenuation per period
coupling_g = g;
b = singleStringZero(note);
a = singleStringPole(note);
tempd = 3*(1-b)-g*(1-a);
b0Coupling = 2*(g*(1-a)-(1-b))/tempd;
b1Coupling = 2*(a*(1-b)-g*(1-a)*b)/tempd;
a1Coupling = (g*(1-a)*b - 3*a*(1-b))/tempd;

//string stiffness
stiffness = stiffnessFactor*stiffnessCoefficient(note);
stiffnessAP = poleZero(b0s,b1s,a1s)
	with{
		b0s = stiffness;
		b1s = 1;
		a1s = stiffness;
	};
// stiffnessAP = poleZero(b0s,b1s,a1s) // Sorta works...
// 	with{
// 		b0s = max(-1, stiffness*TO_SR);
// 		b1s = 1;
// 		a1s = max(-1, stiffness*TO_SR);
// 	};

// uses frequencySmooth only for calculations requiring smoothness.
delayG(frequency,frequencySmooth,stiffnessCoefficient) = de.fdelay(4096*ot.SAMPLE_RATE_TO_IDEAL,delayLength)
	with{
		allPassPhase(a1,WT) = atan2((a1*a1-1.0)*sin(WT),(2.0*a1+(a1*a1+1.0)*cos(WT)));
		poleZeroPhase(b0,b1,a1,WT) = atan2(-b1*sin(WT)*(1 + a1*cos(WT)) + a1*sin(WT)*(b0 + b1*cos(WT)),
						   (b0 + b1*cos(WT))*(1 + a1*cos(WT)) + b1*sin(WT)*a1*sin(WT));
		wT = frequency*2*ma.PI/ma.SR;
		wTSmooth = frequencySmooth*2*ma.PI/ma.SR;
		correction = -1;
		delayLength = (2*ma.PI + 3*allPassPhase(stiffnessCoefficient, wT) +
						poleZeroPhase((1+2*b0Coupling),
						a1Coupling + 2*b1Coupling, a1Coupling, wT)) / wTSmooth + correction;	
	};
		
coupledStrings = (parallelStrings <: (_,(_+_ <: _,_),_ : _,_,(_ : couplingFilter),_ : adder))~(_,_) : !,!,_
	with{
		releaseLoopGain = getValueReleaseLoopGain; // <piano.h>
		getDetuneHz = getValueDetuningHz; // <piano.h>
		heldOrSustained = max(gate, sustain);

		heldDamp = heldOrSustained;//*0.9996; //Is it safe to keep the raw value?
		releasedDamp = (1-heldOrSustained)*releaseLoopGain(note)*0.9;
		// Helps with broken couplingFilter...
		// enhanceSustain = 25/100;
		// noteDamp = it.interpolate_linear(enhanceSustain, coupling_g, 1); //extra sustain
		stringDamp = (heldDamp+releasedDamp);//*noteDamp;

		// attenuating isn't doing what i think it should...?
		//attenuate = *(ba.db2linear(getValueSingleStringDecayRate(note)));


		coupledStringLoopGain = stringDamp; //si.smoo // This used to be smoothed, but I think it's a bad idea...

		couplingFilter = poleZeroSR(b0Coupling,b1Coupling,a1Coupling);
		// Sounds good, but ruins tuning (delayG adjustments):
		// couplingFilter = *(-1) : highpass : lowpass : attenuate with {
		// 	highpass = fi.highpass(1, min(6000*ba.semi2ratio(note_01*12), ot.MAX_FILTER_FREQ));
		// 	lowpass = fi.lowpass(1, min(19000, ot.MAX_FILTER_FREQ));
		// 	attenuate = *(ba.db2linear(note_01 * -30));
		// };
		//couplingFilter = _ <: (couplingFilter_a*(1-ot.input.modWheel), couplingFilter_b*(ot.input.modWheel)) :> _;
		
		detuneHz = getDetuneHz(note);

		tuning = chaos : it.remap(-1, 1, range.low, range.high) with {
			chaos = ot.noiseTable(256, note);
			range = ot.ratioRange(tuningRange-1);
		};

		detuning = chaos : it.remap(-1, 1, range.low, range.high) * detuningFactor with {
			chaos = ot.noiseTable(256, 128+note);
			range = ot.ratioRange((1-tuningFactor)*0.5);
		};

		adjustFreq(f,amt) = f*tuning + amt*detuneHz*detuning;

		freq1 = adjustFreq(freq, 0.5);
		freq2 = adjustFreq(freq, -0.5);
		freq1_smooth = adjustFreq(freqSmooth, 0.5);
		freq2_smooth = adjustFreq(freqSmooth, -0.5);
		delay1 = delayG(freq1,freq1_smooth,stiffness);
		delay2 = delayG(freq2,freq2_smooth,stiffness);
		parallelStrings(x,y) = _ <: (+(x)*coupledStringLoopGain : seq(i,3,stiffnessAP) : delay1),
				(_+y*coupledStringLoopGain : seq(i,3,stiffnessAP) : delay2);
		adder(w,x,y,z) = (y <: +(w),+(z)),x ;	
	};

//==================== PROCESSING ================

// tuningTest(signal) = signal + os.osc(freq) * (signal:an.amp_follower(20/1000)) * 4;

// Main Version
process = soundBoard : *(6) : hammer : dcBlock1 : coupledStrings <: +(eq) : *(ba.db2linear(11)); // : *(ba.db2linear(7));//pianoteq match

// Waveguide Test Version:
//process = soundBoard : *(6) : hammer : dcBlock1 : pianoStrings(freq, 1+detuningFactor/1000) : *(ba.db2linear(0));// : *(ba.db2linear(7));//pianoteq match

// Testing for SampleRate
// process = soundBoard <: (*(6) : hammer : coupledStrings <: +(0)) : *(12) : *(ba.db2linear(-14));


// // Waveguide Attempt:
// //pianoStringSpeakingLength = it.interpolate_linear(sin((pianoNote/88+2.5)*ma.PI)*0.5+0.5, 100, 5);
// pianoString(freq, pos_offset, excite) = pm.endChain(wg)
// with{
//     fc = 20000*ba.semi2ratio((brightnessFactor-1)*12);// hslider("lowpass",20000,10,20000,1);
// 	filter = fi.highshelf(1, -1, fc);// fi.lowpass(1, fc);//si.smooth(1.0-2*(fc/ma.SR));
//     rt = pm.rTermination(pm.basicBlock,*(-1) : filter);
//     lt = pm.lTermination(*(-1)*it.interpolate_linear(gate, 0.65, 0.996), pm.basicBlock);

// 	pos = getValueStrikePosition(note)+pos_offset;
//     length = pm.f2l(freq*2);
//     maxS = ma.SR/20;
// 	lengthS = pm.l2s(length);
//     upS   = lengthS*pos;
//     downS = lengthS - upS;
//     correctionS = -6;
//     wg = pm.chain(lt : pm.waveguide(maxS, upS) : pm.in(excite) : pm.out : pm.waveguide(maxS, downS+correctionS) : rt); // waveguide chain
// };

// pianoStrings(freq, detune, excite) = (string1, string2, string3) :> _ letrec {
//     'string1 = pianoString(freq, -0.01, excite) * enable1;
//     'string2 = pianoString(freq*detune, 0, excite) * enable2;
//     'string3 = pianoString(freq/detune, 0.01, excite) * enable3;
// } with {
// 	// 8 x 1 string
// 	// 12 x 2 string
// 	// 68 x 3 string
// 	// Highest: 5cm speaking length
// 	// Lowest: 100-200cm speaking length
// 	enable1 = pianoNote >= 0;
// 	enable2 = pianoNote >= 8;
// 	enable3 = pianoNote >= 8+12;
// };